#pragma once

#include "v3.h"
#include "framebuffer.h"

class TMesh {
public:
  V3 *verts;
  int vertsN;
  V3 *colors;
  V3 *normals;
  float *sts;
  FrameBuffer *texture;
  unsigned int *tris;
  int trisN;
  M3x3 *triMs;
  // triangle t has vertices 
  // (verts[tris[3*t+0]], verts[tris[3*t+1]], verts[tris[3*t+2]])
  TMesh();
  void SetBaseRectangle(V3 center, float dx, float dz);
  void Load(char *fname);
  void RenderWF(FrameBuffer *fb, PPC *ppc, unsigned int wfc);
  void RenderFilled(FrameBuffer *fb, 
    PPC *ppc, V3 lightPoint, 
    FrameBuffer *shadowMap, PPC *lightCamera,
    bool msi);
  void RenderRT(FrameBuffer *fb, PPC *ppc);
  void Translate(V3 trans);
  void Scale(float diag);
  V3 GetCenter();
  AABB GetAABB();
  void SetCenter(V3 newCenter);
  void SetColors(V3 color);
  bool Clip(V3 *projVerts, int *vi, int w, int h,
    int& left, int& right, int &top, int &bottom);
  M3x3 SetABCMatrix(M3x3 ssim, V3 r0, V3 r1, V3 r2);
  M3x3 SetMSIQ(V3 V0, V3 V1, V3 V2, PPC *ppc);
  V3 SetDenABC(M3x3 Q);
  V3 SetNumABC(M3x3 Q, V3 r);
  void SetFromFrambuffer(FrameBuffer *fb, PPC *ppc);
  bool Intersect(V3 vs[3], V3 rayO, V3 rayD, V3 &iP, float &w);

  void RenderHW();
  void SetTriMs();

};